unit GWDragSource;
{---------------------------------------------------------------
Purpose: Implements essential foundation for OLE Drag Source.

Author:  Graham Wideman
         graham@wideman-one.com
         www.wideman-one.com

TGWDragAppInterface
--------------------
Implements:
1) monitoring functions
2) default dumb implementations for IDropSource event Callbacks
   Override this class to provide app-specific functions

TGWDropSource, TGWDragDataObject, TGWEnumFormatEtc.
----------------------------------------------------
Implements:
1)  generic interfaces.
2)  delegates significant functions to TGWDragAppInterface
3)  generic monitoring for testing
Should be useable as-is in apps... place all app-specific functionality in
descendant of TGWDragAppInterface

Revision History
----------------
98-07-20 GW V1.1 original (version sync'ed with GWDropTarget)

To-do list
----------

---------------------------------------------------------------}

//-----------------------
interface
//-----------------------
uses Windows, Messages, SysUtils, Classes, graphics,
     activex, stdctrls, controls, forms, grids,
     GWMonOLE;

Type
  //---------------------------------------
  // forward declarations
  //---------------------------------------
  TGWDropSource           = class;
  TGWDragDataObject       = class;
  TGWDragEnumFormatEtc    = class;

  //---------------------------------------
  TGWDragAppInterface = class
  //---------------------------------------
  protected
    FOwnerForm         : TForm;
    Ffm_OLEMonitor     : Tfm_OLEMonitor;

  public
    property fm_OLEMonitor: Tfm_OLEMonitor  read Ffm_OLEMonitor;

    //---------------------------------------
    // Production functions
    //---------------------------------------
    // Create in  OwnerForm.OnCreate, Destroy in OwnerForm.OnDestroy
    constructor Create(AOwner: TForm; UseMonitor: Boolean); virtual;
    function    DoDragDrop(dwOKEffects: Longint; var dwEffect: Longint): HResult; virtual;

  protected
    //---------------------------------------
    // delegated functions
    //---------------------------------------
    //TGWDropSource
    function QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult; virtual;
    function GiveFeedback(dwEffect: Longint): HResult; virtual;

    //TGWDragDataObject
    function QueryGetData(const formatetc: TFormatEtc): HResult; virtual;
    function GetData     (const formatetc: TFormatEtc; var medium: TStgMedium): HResult; virtual;

    //TGWDragEnumFormatEtc
    function FormatCount: integer; virtual;
    procedure FormatEtc(N: integer; var formatetc: TFormatEtc); virtual;

  public


  end;

  //------------------------------------
  TGWDropSource = class (TGWMonInterfacedObject, IDropSource)
  //------------------------------------
  protected
    //-----------------------------
    // Monitor and delegate
    //-----------------------------
    FDragAppInterface  : TGWDragAppInterface;  // DragAppInterface should nil this on destroy

  public
    constructor Create(ADragAppInterface: TGWDragAppInterface); virtual;

  public
    //---------------------------------
    //   IDropSource functions
    //---------------------------------
    function QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult; stdcall;
    function GiveFeedback(dwEffect: Longint): HResult; stdcall;
  end;

  //------------------------------------
  TGWDragDataObject = class (TGWMonInterfacedObject, IDataObject)
  //------------------------------------
  protected
    //-----------------------------
    // Monitor and delegate
    //-----------------------------
    FDragAppInterface  : TGWDragAppInterface;  // DragAppInterface should nil this on destroy

  public
    constructor Create(ADragAppInterface: TGWDragAppInterface); virtual;

  public
    //---------------------------------
    //   IDataObject functions
    //---------------------------------
    function GetData     (const formatetc: TFormatEtc; out medium: TStgMedium): HResult; stdcall;
    function GetDataHere (const formatetc: TFormatEtc; out medium: TStgMedium): HResult; stdcall;
    function QueryGetData(const formatetc: TFormatEtc): HResult; stdcall;
    function SetData     (const formatetc: TFormatEtc; var medium: TStgMedium; fRelease: BOOL): HResult; stdcall;
    function GetCanonicalFormatEtc(const formatetcIn: TFormatEtc;   out formatetcOut: TFormatEtc): HResult; stdcall;
    function EnumFormatEtc(dwDirection: Longint; out enumFormatEtc: IEnumFormatEtc): HResult; stdcall;
    function DAdvise(const formatetc: TFormatEtc; advf: Longint; const advSink: IAdviseSink; out dwConnection: Longint): HResult; stdcall;
    function DUnadvise(dwConnection: Longint): HResult; stdcall;
    function EnumDAdvise(out enumAdvise: IEnumStatData): HResult; stdcall;

  end;

  //------------------------------------
  TGWDragEnumFormatEtc = class (TGWMonInterfacedObject, IEnumFormatEtc)
  //------------------------------------
  protected
    //-----------------------------
    // Monitor and delegate
    //-----------------------------
    FIndex             : integer;              // Next FormatEtc to get
    FDragAppInterface  : TGWDragAppInterface;  // DragAppInterface should nil this on destroy

  public
    constructor Create(ADragAppInterface: TGWDragAppInterface); virtual;
    destructor  Destroy; override;

  public
    //---------------------------------
    //   IEnumFormatEtc functions
    //---------------------------------
    function Next (CountRequested: Longint; out FormatEtcArray; PCountFetched: PLongint): HResult; stdcall;
    function Skip (count: Longint) : HResult; stdcall;
    function Reset : HResult; stdcall;
    function Clone (out enum : IEnumFormatEtc) : HResult; stdcall;
  end;

  TFormatEtcArray = array[0..19] of TFormatEtc;
  PFormatEtcArray = ^TFormatEtcArray;

//-----------------------
implementation
//-----------------------

//-------------------------------------------------------
//        TGWDragEnumFormatEtc
//-------------------------------------------------------
Const
  // data items
  di_Next   = gwmd_BeginObjectSpecific;
  di_Skip   = gwmd_BeginObjectSpecific + 1;
  di_Reset  = gwmd_BeginObjectSpecific + 2;
  di_Clone  = gwmd_BeginObjectSpecific + 3;

//-------------------------------------------------------
constructor TGWDragEnumFormatEtc.Create(ADragAppInterface: TGWDragAppInterface);
//-------------------------------------------------------
Begin
  inherited Create(ADragAppInterface.Ffm_OLEMonitor);
  FIndex            := 0;
  FDragAppInterface := ADragAppInterface;

  If Assigned(FMon) then
  Begin
    FMon.Monitoree_AddData('Next ');
    FMon.Monitoree_AddData('Skip ');
    FMon.Monitoree_AddData('Reset');
    FMon.Monitoree_AddData('Clone');
  end;
end;

//-------------------------------------------------------
destructor  TGWDragEnumFormatEtc.Destroy;
//-------------------------------------------------------
Begin

  inherited Destroy;
end;

//-------------------------------------------------------
function TGWDragEnumFormatEtc.Next (CountRequested: Longint; out FormatEtcArray; PCountFetched: PLongint): HResult;
//-------------------------------------------------------
Var
  N: integer;
  FormatEtcArrayOut: TFormatEtcArray absolute FormatEtcArray;
Label the_end;
Begin
  Result := S_FALSE;
  if not Assigned(FDragAppInterface) then goto the_end; // app or form no longer around;

  N := 0;
  while (N < CountRequested) and (FIndex < FDragAppInterface.FormatCount) do
  Begin
    FDragAppInterface.FormatEtc(FIndex, FormatEtcArrayOut[N]);
    Inc(FIndex);
    Inc(N);
  end;
  If PCountFetched <> nil then PCountFetched^ := N;
  if N = CountRequested then Result := S_OK;

the_end:
  Mon_IncData(di_Next );
end;

//-------------------------------------------------------
function TGWDragEnumFormatEtc.Skip (count: Longint) : HResult;
//-------------------------------------------------------
Begin
  Result := S_FALSE;
  If Assigned(FDragAppInterface) then
  Begin
    FIndex := FIndex + count;
    If FIndex > FDragAppInterface.FormatCount then
    Begin
      FIndex := FDragAppInterface.FormatCount;
      Result := S_FALSE;
    end else Result := S_OK;
  end;
  Mon_IncData(di_Skip );
end;

//-------------------------------------------------------
function TGWDragEnumFormatEtc.Reset : HResult;
//-------------------------------------------------------
Begin
  FIndex := 0;
  Result := S_OK;
  Mon_IncData(di_Reset);
end;

//-------------------------------------------------------
function TGWDragEnumFormatEtc.Clone (out enum : IEnumFormatEtc) : HResult; stdcall;
//-------------------------------------------------------
Var
  MyFormatEtc: TGWDragEnumFormatEtc;
Begin
  MyFormatEtc := TGWDragEnumFormatEtc.Create(FDragAppInterface);
  enum := MyFormatEtc;
  Result := S_OK;
  Mon_IncData(di_Clone);
end;

//-------------------------------------------------------
//        TGWDragDataObject
//-------------------------------------------------------
Const
  // data items
  di_GetData                = gwmd_BeginObjectSpecific;
  di_GetDataHere            = gwmd_BeginObjectSpecific + 1;
  di_SetData                = gwmd_BeginObjectSpecific + 2;
  di_QueryGetData           = gwmd_BeginObjectSpecific + 3;
  di_GetCanonicalFormatEtc  = gwmd_BeginObjectSpecific + 4;
  di_EnumFormatEtc          = gwmd_BeginObjectSpecific + 5;
  di_DAdvise                = gwmd_BeginObjectSpecific + 6;
  di_DUnadvise              = gwmd_BeginObjectSpecific + 7;
  di_EnumDAdvise            = gwmd_BeginObjectSpecific + 8;

//-------------------------------------------------------
constructor TGWDragDataObject.Create(ADragAppInterface: TGWDragAppInterface);
//-------------------------------------------------------
Begin
  inherited Create(ADragAppInterface.Ffm_OLEMonitor);
  FDragAppInterface  := ADragAppInterface;

  If Assigned(FMon) then
  Begin
    FMon.Monitoree_AddData('GetData');
    FMon.Monitoree_AddData('GetDataHere');
    FMon.Monitoree_AddData('SetData');
    FMon.Monitoree_AddData('QueryGetData');
    FMon.Monitoree_AddData('GetCanonicalFormatEtc');
    FMon.Monitoree_AddData('EnumFormatEtc');
    FMon.Monitoree_AddData('DAdvise');
    FMon.Monitoree_AddData('DUnadvise');
    FMon.Monitoree_AddData('EnumDAdvise');
  end;
end;

//-------------------------------------------------------
function TGWDragDataObject.EnumFormatEtc(dwDirection: Longint; out enumFormatEtc: IEnumFormatEtc): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Mon_IncData(di_EnumFormatEtc);

  if dwDirection <> DATADIR_GET then
  Begin
    enumFormatEtc := nil;
    Result := E_NOTIMPL;
    goto the_end;
  end;
  //--------------------------------------------------------------------
  // Finer OLE point: assigning to enumFormatEtc will auto increment the
  // RefCount to 1, since enumFormatEtc is an IEnumFormatEtc.
  // ie: no need to _AddRef, in fact that will lead to enumFormatEtc
  // never getting deleted.
  // Note, this conflicts with Cantu 1998 Delphi Developer Handbook P465
  //--------------------------------------------------------------------
  enumFormatEtc := TGWDragEnumFormatEtc.Create(FDragAppInterface);
  Result := S_OK;
the_end:
end;

//-------------------------------------------------------
function TGWDragDataObject.QueryGetData(const formatetc: TFormatEtc): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Result := DV_E_FORMATETC;
  If (not Assigned(FDragAppInterface)) then goto the_end;
  Result := FDragAppInterface.QueryGetData(formatetc);
the_end:
  Mon_IncData(di_QueryGetData         );
end;

//-------------------------------------------------------
function TGWDragDataObject.GetData(const formatetc: TFormatEtc; out medium: TStgMedium): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Result               := DV_E_FORMATETC;
  medium.tymed         := 0;
  medium.hGlobal       := 0;
  medium.unkForRelease := nil;
  If (FDragAppInterface = nil) then goto the_end;
  Result := FDragAppInterface.GetData(formatetc, medium);
the_end:
  Mon_IncData(di_GetData              );
end;

//-------------------------------------------------------
function TGWDragDataObject.GetDataHere (const formatetc: TFormatEtc; out medium: TStgMedium): HResult;
//-------------------------------------------------------
Begin
  Result := DV_E_FORMATETC;
  Mon_IncData(di_GetDataHere          );
end;

//-------------------------------------------------------
function TGWDragDataObject.SetData(const formatetc: TFormatEtc; var medium: TStgMedium; fRelease: BOOL): HResult;
//-------------------------------------------------------
Begin
  Result := DV_E_FORMATETC;
  Mon_IncData(di_SetData              );
end;

//-------------------------------------------------------
function TGWDragDataObject.GetCanonicalFormatEtc(const formatetcIn: TFormatEtc;   out formatetcOut: TFormatEtc): HResult;
//-------------------------------------------------------
Begin
  formatetcOut.ptd := nil;
  Result := E_NOTIMPL;
  Mon_IncData(di_GetCanonicalFormatEtc);
end;

//-------------------------------------------------------
function TGWDragDataObject.DAdvise(const formatetc: TFormatEtc; advf: Longint; const advSink: IAdviseSink; out dwConnection: Longint): HResult; stdcall;
//-------------------------------------------------------
Begin
  Result := OLE_E_ADVISENOTSUPPORTED;
  Mon_IncData(di_DAdvise              );
end;

//-------------------------------------------------------
function TGWDragDataObject.DUnadvise(dwConnection: Longint): HResult; stdcall;
//-------------------------------------------------------
Begin
  Result := OLE_E_ADVISENOTSUPPORTED;
  Mon_IncData(di_DUnadvise            );
end;

//-------------------------------------------------------
function TGWDragDataObject.EnumDAdvise(out enumAdvise: IEnumStatData): HResult; stdcall;
//-------------------------------------------------------
Begin
  Result := OLE_E_ADVISENOTSUPPORTED;
  Mon_IncData(di_EnumDAdvise          );
end;

//----------------------------------------------------------
//      TGWDropSource
//----------------------------------------------------------
Const
  // data items
  di_QueryContinueDrag = gwmd_BeginObjectSpecific;
  di_GiveFeedback      = gwmd_BeginObjectSpecific + 1;

//-------------------------------------------------------
constructor TGWDropSource.Create(ADragAppInterface: TGWDragAppInterface);
//-------------------------------------------------------
Begin
  inherited Create(ADragAppInterface.Ffm_OLEMonitor);
  FDragAppInterface  := ADragAppInterface;

  If Assigned(FMon) then
  Begin
    FMon.Monitoree_AddData('QueryContinueDrag');
    FMon.Monitoree_AddData('GiveFeedback');
  end;
end;

//-------------------------------------------------------
function TGWDropSource.QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  If (FDragAppInterface = nil) then goto the_end;
  Result := FDragAppInterface.QueryContinueDrag(fEscapePressed, grfKeyState);
the_end:
  Mon_IncData(di_QueryContinueDrag);
end;

//------------------------------------------------
function TGWDropSource.GiveFeedback(dwEffect: Longint): HResult;
//------------------------------------------------
Label the_end;
Begin
  If (FDragAppInterface = nil) then goto the_end;
  Result := FDragAppInterface.GiveFeedback(dwEffect);
the_end:
  Mon_IncData(di_GiveFeedback);
end;

//-------------------------------------------------------
//     TGWDragAppInterface
//-------------------------------------------------------

//-----------------------------------------
constructor TGWDragAppInterface.Create(AOwner: TForm;  UseMonitor: Boolean );
//-----------------------------------------
Begin
  inherited Create;
  FOwnerForm         := AOwner;
  If not UseMonitor then Ffm_OLEMonitor := nil else
  Begin
    Ffm_OLEMonitor := Tfm_OLEMonitor.Create(FOwnerForm);
  end;
end;

//---------------------------------------
function TGWDragAppInterface.DoDragDrop(dwOKEffects: Longint; var dwEffect: Longint): HResult;
//---------------------------------------
Var
  DropSource   : TGWDropSource;      // Pascal versions for debugging purposes
  DataObject   : TGWDragDataObject;
  MyIDropSource: IDropSource;        // Ixxx versions to generate references
  MyIDataObject: IDataObject;
  N: integer;
Begin
  //---------------------------------------------------------------
  // Finer OLE point: There are tricks to getting reference counting
  // to work properly (and automatically)
  //
  // Note: Following is contrary to Mischel in "High Performance
  // Delphi 3 Programming, page 126.
  //
  // Strategy 1:
  // -----------
  //  DoDragDrop(..., FDropSource,...);
  //  DOESN'T WORK PROPERLY
  //  ... FDropSource is a TGWDropSOurce. This *appears* to work OK
  //  but doesn't increment RefCount, and the OS doesn't either,
  //  so object never gets deleted.
  //
  // Strategy 2:
  // -----------
  //  DoDragDrop(..., FDropSource as IDropSource,...);
  // ... this would appear to be merely making explicit what Strategy 1
  //  does implicitly, however, this DOES increment the reference count
  // properly, and decrement it on return from call. Must generate temporary
  // IDropSource variable for duration of call.
  //
  // Strategy 3:
  // -----------
  // MyIDropSource    := FDropSource; { <-- increments RefCount }
  // ...
  // DoDragDrop(..., MyIDropSource,...);
  //
  // ... works correctly. Later, at procedure "end;",  object gets released
  // as MyIDropSource goes out of scope. (set breakpoint on "end;" to test this).
  //---------------------------------------------------------------

  DropSource     := TGWDropSource.Create(self);
  MyIDropSource  := DropSource;

  DataObject     := TGWDragDataObject.Create(self);
  MyIDataObject  := DataObject;

  Result := activex.DoDragDrop(MyIDataObject, MyIDropSource, dwOKEffects, dwEffect);
end;

//---------------------------------------
// TGWDragAppInterface delegated functions
//---------------------------------------

//TGWDropSource:
//-------------------------------------------------------
function TGWDragAppInterface.QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult;
//-------------------------------------------------------
Begin
  Result := S_OK;
  If fEscapePressed then Result := DRAGDROP_S_CANCEL else
    if ((grfKeyState and MK_LBUTTON) = 0) then // mouse-up
      Result := DRAGDROP_S_DROP;
end;

//------------------------------------------------
function TGWDragAppInterface.GiveFeedback(dwEffect: Longint): HResult;
//------------------------------------------------
Begin
  Result := DRAGDROP_S_USEDEFAULTCURSORS;
end;

//TGWDragDataObject
//---------------------------------------
function TGWDragAppInterface.QueryGetData(const formatetc: TFormatEtc): HResult;
//---------------------------------------
Begin
  Result := DV_E_FORMATETC;
  if (formatetc.dwAspect = DVASPECT_CONTENT) and
     (formatetc.cfFormat = CF_TEXT) and
     (formatetc.tymed = TYMED_HGLOBAL) then Result := S_OK;
end;

//---------------------------------------
function TGWDragAppInterface.GetData(const formatetc: TFormatEtc; var medium: TStgMedium): HResult;
//---------------------------------------
Var
  HGlobalData: HGlobal;
  PGlobalData: PChar;
  DemoString: string;
  DemoStringLen: integer;
  AllocLen: integer;

Label the_end;
Begin
  Result := DV_E_FORMATETC;
  If QueryGetData(formatetc) <> S_OK then goto the_end;

  // most medium fields are already initialized by this point
  medium.tymed := TYMED_HGLOBAL;

  DemoString     := 'Default demo string';
  DemoStringLen  := Length(DemoString);
  AllocLen := DemoStringLen + 1;

  HGlobalData := GlobalAlloc((GMEM_SHARE or GMEM_MOVEABLE or GMEM_ZEROINIT), AllocLen);
  if (HGlobalData <> 0) then
  begin
    PGlobalData := GlobalLock (HGlobalData); // lock while we are using it
    StrCopy(PGlobalData, PChar(DemoString));
    GlobalUnlock (HGlobalData);
    medium.hGlobal := HGlobalData;
    Result := S_OK;
  end;

the_end:
end;

//TGWDragEnumFormatEtc
//---------------------------------------
function TGWDragAppInterface.FormatCount: integer;
//---------------------------------------
Begin // dumb default implementation
  Result := 1;
end;

//---------------------------------------
procedure TGWDragAppInterface.FormatEtc(N: integer; var formatetc: TFormatEtc);
//---------------------------------------
Begin // dumb default implementation
  // ignore N, always say we do CF_TEXT
  formatetc.cfFormat := CF_TEXT;
  formatetc.ptd      := nil;
  formatetc.dwAspect := DVASPECT_CONTENT;
  formatetc.lindex   := -1;
  formatetc.tymed    := TYMED_HGLOBAL;
end;

//-----------------------
initialization
//-----------------------
  OleInitialize (Nil);

//-----------------------
finalization
//-----------------------
  OleUninitialize;
end.

