unit GWCOMDemo_DoDragDrop;
{---------------------------------------------------------------
Purpose: Quick-and-dirty unit that demos rudimentary
         DoDragDrop call showing RefCount anomaly.

Author:  Graham Wideman
         graham@wideman-one.com
         www.wideman-one.com

Revision History
----------------
98-07-26 GW V1.1 original

To-do list
----------

---------------------------------------------------------------}

//-----------------------
interface
//-----------------------
uses Windows, Messages, SysUtils, Classes, ActiveX, forms,
     GWMonOLE, GWCOMDemo_Base;

Type
  //-------------------------------
  TGWCOMDemo_DoDragDrop = class(TGWCOMDemo_Base)
  //-------------------------------
  public
    procedure Demo_Run_Inner_0;
    procedure Demo_Run_Inner_1;
    procedure Demo_Run_Inner_2;
    procedure Demo_Run; override;
  end;

  //-----------------------------------------------
  // A very rudimentary implementation of
  // Drag objects just to demo DoDragDrop call
  //-----------------------------------------------

  //------------------------------------
  TGWDropSource = class (TGWMonInterfacedObject, IDropSource)
  //------------------------------------
  public
    constructor Create(Afm_OLEMonitor: Tfm_OLEMonitor); virtual;

  public
    //---------------------------------
    //   IDropSource functions
    //---------------------------------
    function QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult; stdcall;
    function GiveFeedback(dwEffect: Longint): HResult; stdcall;
  end;

  //------------------------------------
  TGWDragDataObject = class (TGWMonInterfacedObject, IDataObject)
  //------------------------------------
  public
    constructor Create(Afm_OLEMonitor: Tfm_OLEMonitor); virtual;
  public
    //---------------------------------
    //   IDataObject functions
    //---------------------------------
    function GetData     (const formatetc: TFormatEtc; out medium: TStgMedium): HResult; stdcall;
    function GetDataHere (const formatetc: TFormatEtc; out medium: TStgMedium): HResult; stdcall;
    function QueryGetData(const formatetc: TFormatEtc): HResult; stdcall;
    function SetData     (const formatetc: TFormatEtc; var medium: TStgMedium; fRelease: BOOL): HResult; stdcall;
    function GetCanonicalFormatEtc(const formatetcIn: TFormatEtc;   out formatetcOut: TFormatEtc): HResult; stdcall;
    function EnumFormatEtc(dwDirection: Longint; out enumFormatEtc: IEnumFormatEtc): HResult; stdcall;
    function DAdvise(const formatetc: TFormatEtc; advf: Longint; const advSink: IAdviseSink; out dwConnection: Longint): HResult; stdcall;
    function DUnadvise(dwConnection: Longint): HResult; stdcall;
    function EnumDAdvise(out enumAdvise: IEnumStatData): HResult; stdcall;
  end;

//-----------------------
implementation
//-----------------------
uses GWDemoCode;

//-------------------------------------------------------
//        TGWDragDataObject
//-------------------------------------------------------

//-------------------------------------------------------
constructor TGWDragDataObject.Create(Afm_OLEMonitor: Tfm_OLEMonitor);
//-------------------------------------------------------
Begin
  inherited Create(Afm_OLEMonitor);
end;

//-------------------------------------------------------
function TGWDragDataObject.EnumFormatEtc(dwDirection: Longint; out enumFormatEtc: IEnumFormatEtc): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Result := E_NOTIMPL;
  Result := S_OK;
the_end:
end;

//-------------------------------------------------------
function TGWDragDataObject.QueryGetData(const formatetc: TFormatEtc): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Result := DV_E_FORMATETC;
end;

//-------------------------------------------------------
function TGWDragDataObject.GetData(const formatetc: TFormatEtc; out medium: TStgMedium): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Result               := DV_E_FORMATETC;
end;

//-------------------------------------------------------
function TGWDragDataObject.GetDataHere (const formatetc: TFormatEtc; out medium: TStgMedium): HResult;
//-------------------------------------------------------
Begin
  Result := DV_E_FORMATETC;
end;

//-------------------------------------------------------
function TGWDragDataObject.SetData(const formatetc: TFormatEtc; var medium: TStgMedium; fRelease: BOOL): HResult;
//-------------------------------------------------------
Begin
  Result := DV_E_FORMATETC;
end;

//-------------------------------------------------------
function TGWDragDataObject.GetCanonicalFormatEtc(const formatetcIn: TFormatEtc;   out formatetcOut: TFormatEtc): HResult;
//-------------------------------------------------------
Begin
  formatetcOut.ptd := nil;
  Result := E_NOTIMPL;
end;

//-------------------------------------------------------
function TGWDragDataObject.DAdvise(const formatetc: TFormatEtc; advf: Longint; const advSink: IAdviseSink; out dwConnection: Longint): HResult; stdcall;
//-------------------------------------------------------
Begin
  Result := OLE_E_ADVISENOTSUPPORTED;
end;

//-------------------------------------------------------
function TGWDragDataObject.DUnadvise(dwConnection: Longint): HResult; stdcall;
//-------------------------------------------------------
Begin
  Result := OLE_E_ADVISENOTSUPPORTED;
end;

//-------------------------------------------------------
function TGWDragDataObject.EnumDAdvise(out enumAdvise: IEnumStatData): HResult; stdcall;
//-------------------------------------------------------
Begin
  Result := OLE_E_ADVISENOTSUPPORTED;
end;

//----------------------------------------------------------
//      TGWDropSource
//----------------------------------------------------------

//-------------------------------------------------------
constructor TGWDropSource.Create(Afm_OLEMonitor: Tfm_OLEMonitor);
//-------------------------------------------------------
Begin
  inherited Create(Afm_OLEMonitor);
end;

//-------------------------------------------------------
function TGWDropSource.QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult;
//-------------------------------------------------------
Label the_end;
Begin
  Result := S_OK;
  If fEscapePressed then Result := DRAGDROP_S_CANCEL else
    if ((grfKeyState and MK_LBUTTON) = 0) then // mouse-up
      Result := DRAGDROP_S_DROP;
end;

//------------------------------------------------
function TGWDropSource.GiveFeedback(dwEffect: Longint): HResult;
//------------------------------------------------
Label the_end;
Begin
  Result := DRAGDROP_S_USEDEFAULTCURSORS;
end;

//-------------------------------------------------------
//     TGWDragAppInterface
//-------------------------------------------------------

//ShowBegin
// A set of demos showing different calls to DoDragDrop, and
// what happens to RefCount in each case

//-----------------------------------------------
procedure TGWCOMDemo_DoDragDrop.Demo_Run_Inner_0;
//-----------------------------------------------
Var
  DropSource   : TGWDropSource;
  DataObject   : TGWDragDataObject;
  Result: HResult;
  dwOKEffects, dwEffect: longint;
Begin
  Mon.ClearDeadMonitors;
  Pause('DDD-000');
  dwOKEffects    := DROPEFFECT_COPY;

  DropSource     := TGWDropSource.Create(Mon);
  DataObject     := TGWDragDataObject.Create(Mon);
  Pause('DDD-020');

  //------------------------------------------------
  // Try calling DoDragDrop with Pascal Object
  // NOTE: THIS IS THE WRONG WAY!
  //------------------------------------------------
  Result := DoDragDrop(DataObject, DropSource, dwOKEffects, dwEffect);
  Pause('DDD-030'); // Note DropSource RefCount never got incremented

  {-----------------------------------------------------
   If RefCount never gets incremented, then the object won't
   ever get deleted (unless we explicitly call Destroy).
   The objects are usable by the called function, but the
   RefCount does not get incremented.
   It's the caller's responsibility to "allocate the reference"
   that it passes (ie: increment the RefCount), so this appears
   to be one place where Delphi doesn't do quite what we
   need automatically.  But no matter, as there are two
   other preferable ways to make this call... as we will see.
  ---------------------------------------------------------}

  if Assigned(DropSource) then DropSource.Destroy; // fix up problem
  Pause('DDD-040');
end;

//-----------------------------------------------
procedure TGWCOMDemo_DoDragDrop.Demo_Run_Inner_1;
//-----------------------------------------------
Var
  DropSource   : TGWDropSource;
  DataObject   : TGWDragDataObject;
  Result: HResult;
  dwOKEffects, dwEffect: longint;
Begin
  Mon.ClearDeadMonitors;
  Pause('DDD-100');
  dwOKEffects    := DROPEFFECT_COPY;

  //-------------------------------------------------
  // Try typecasting the objects explicitly
  //-------------------------------------------------
  DropSource     := TGWDropSource.Create(Mon);
  DataObject     := TGWDragDataObject.Create(Mon);
  Pause('DDD-120');

  Result := DoDragDrop(DataObject as IDataObject,
                       DropSource as IDropSource,
                       dwOKEffects, dwEffect);
  Pause('DDD-140'); // Note both RefCounts got incremented...
                    // temporary objects will be destroyed at end of scope
end;

//-----------------------------------------------
procedure TGWCOMDemo_DoDragDrop.Demo_Run_Inner_2;
//-----------------------------------------------
Var
  DropSource   : TGWDropSource;
  DataObject   : TGWDragDataObject;
  MyIDropSource: IDropSource;        // Ixxx versions to generate references
  MyIDataObject: IDataObject;
  Result: HResult;
  dwOKEffects, dwEffect: longint;
Begin
  Mon.ClearDeadMonitors;
  Pause('DDD-200');
  dwOKEffects    := DROPEFFECT_COPY;

  //-------------------------------------------------
  // Create actual local COM references and use those
  //-------------------------------------------------
  DropSource     := TGWDropSource.Create(Mon);
  DataObject     := TGWDragDataObject.Create(Mon);
  MyIDropSource  := DropSource; // as IDropSource is optional?
  MyIDataObject  := DataObject; // as IDataObject is optional?

  Pause('DDD-220');
  Result := DoDragDrop(MyIDataObject, MyIDropSource, dwOKEffects, dwEffect);

  Pause('DDD-230'); // Note objects still exist and MyIxxx refs are still good
                   // as expected. Objects will be delete at end of scope.
end;

//---------------------------------------
procedure TGWCOMDemo_DoDragDrop.Demo_Run;
//---------------------------------------
Begin
  Pause('Begin');
  Demo_Run_Inner_0;       Pause('DR-000');  // objects deleted?
  Demo_Run_Inner_1;       Pause('DR-010');  // objects deleted?
  Demo_Run_Inner_2;       Pause('End');     // objects deleted?
end;
//ShowEnd


//-----------------------
initialization
//-----------------------
  OleInitialize (Nil);
  DemoCode_Mgr.Add(TGWCOMDemo_DoDragDrop.Create('DoDragDrop call anomaly','GWCOMDemo_DoDragDrop.pas'));
//-----------------------
finalization
//-----------------------
  OleUninitialize;

end.

